' Microchip PICBASIC PRO Firmware - RS-485 network demo

' FileName:        rs485.bas
' Processor:       PIC18
' Compiler:        PICBASIC PRO

' Software License Agreement
'
' Licensor grants any person obtaining a copy of this software ("You") 
' a worldwide, royalty-free, non-exclusive license, for the duration of 
' the copyright, free of charge, to store and execute the Software in a 
' computer system and to incorporate the Software or any portion of it 
' in computer programs You write.   

' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
' IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
' FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
' AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
' LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
' OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
' THE SOFTWARE.

' Author               Date        Comment
' Jan Axelson          10/20/07    Original

'************************************************************************

' The PIC is a secondary node in an RS-485 network. The USART interfaces 
' to an RS-485 transceiver. A network can have multiple secondary nodes.

' The PIC responds to defined commands. Each command has this format:
' byte 1: ":"
' byte 2: the PIC's network address
' byte 3: command code
' bytes 4..n: additional data (command-dependent)
' final byte: LF code (0ah)

' Each response has this format:
' byte 1: ":"
' byte 2: the PIC's network address
' byte 3..n: additional data (command dependent)
' final byte: LF code (0ah)

' This example defines two commands, read_byte and write_byte.

' To read a byte, send this command to the PIC:

' :h2b 

' followed by a LF code (hit Enter) 

' The PIC will respond with:

' :hxx

' followed by a LF, where xx is the hex value read.

' To write a byte, send this command to the PIC:

' :h1bxx 

' followed by a LF, where xx is the hex value to write.
' (For example, :h1bff or :h1b00) 

' The PIC will respond with:

' :h

' followed by a LF.

' A PC can communicate with the PIC using a custom application, or you can use
' a terminal-emulator such as HyperTerminal if the PC's driver-enable line is 
' controlled entirely in hardware.

' On the PIC, firmware can control the RS-485 driver-enable line if needed. 

' This application and host applications to communicate with the device are available
' from www.Lvr.com.

'************************************************************************

' Serial port configuration

DEFINE HSER_RCSTA 90h
DEFINE HSER_TXSTA 24h
DEFINE HSER_SPBRG 19h
DEFINE HSER_CLROERR

COMMAND_START		con	":"
MAX_COMMAND_LENGTH	con	5
MY_ADDRESS  		con	"h"

' Use any spare output port bit:

driver_enable			var	PORTB.3  'UNNEEDED IF USING HARDWARE-ONLY DRIVER ENABLE	

command_index			var	byte
command_response		var	byte[6]
converted_byte			var	byte
data_ready_to_send		var	bit
delay_before_responding var bit
firmware_driver_enable	var	bit
index 					var byte
lower_nibble			var	byte
network_state 			var byte
received_command		var	byte[6]
response_index			var	byte
serial_in				var	byte
serial_out 				var byte
success					var	bit
upper_nibble			var	byte
value_to_convert		var	byte

command_index = 0
command_response[0] = COMMAND_START
command_response[1] = MY_ADDRESS
network_state = "r"
response_index = 0

' Output port bit used in testing (write_byte command):

TRISB.0 = 0
low PORTB.0

' Uncomment one of these lines:

'delay_before_responding = 0  ' No delay before responding.
delay_before_responding = 1  ' Delay before responding

' Uncomment one of these lines:

firmware_driver_enable = 0  ' Circuits have automatic driver-enable control.
'firmware_driver_enable = 1  ' Firmware must control the driver enable.

if (firmware_driver_enable = 1) then

	' Configure the driver-enable line and disable the driver.

	TRISB.3 = 0
	driver_enable = 0

endif

loop:

	' The main program loop

	gosub serial_communications

	' Add other tasks here.

goto loop

'************************************************************************
' Routine:        	ascii_hex_to_byte 

' PreCondition:    	None

' Input:           	upper_nibble - ASCII hex code for the upper 4 bits of a byte
'					lower_nibble - ASCII hex code for the lower 4 bits of a byte

' Output:          	converted_byte - the value represented by upper_nibble 
'					and lower_nibble
'					success - indicates conversion success (1) or failure (0)
'
' Side Effects:    	None

' Overview:        	Converts 2 ASCII hex characters
'					to the byte value they represent.
                  
' Note:            	Assumes hex characters a-f are lower case.
'************************************************************************

ascii_hex_to_byte:

	success = 1

	' Convert each character code to the value it represents.

	select case upper_nibble

		case "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
			upper_nibble = upper_nibble - 48

		case "a", "b", "c", "d", "e", "f" 
			upper_nibble = upper_nibble - 87

		case "A", "B", "C", "D", "E", "F" 
			upper_nibble = upper_nibble - 55

		case else

			' The text character isn't 0-9, a-f, or A-F.

			success = 0
	end select

	select case lower_nibble

		case "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
			lower_nibble = lower_nibble - 48

		case "a", "b", "c", "d", "e", "f" 
			lower_nibble = lower_nibble - 87

		case "A", "B", "C", "D", "E", "F" 
			lower_nibble = lower_nibble - 55

		case else

			' The text character isn't 0-9, a-f, or A-F.

			success = 0
	end select

	if (success = 1) then

		' Combine the nibbles in a byte.

		converted_byte = (upper_nibble << 4) + lower_nibble
	endif

return

'************************************************************************
' Routine:        	byte_to_ascii_hex 

' PreCondition:    	None

' Input:           	value_to_convert - the byte value to convert

' Output:          	upper_nibble - character code representing the upper 4 bits
'					lower_nibble - character code representing the lower 4 bits

' Side Effects:    	None

' Overview:        	Convert a byte to 2 ASCII hex characters
                  
' Note:            	
'************************************************************************

byte_to_ascii_hex:

	upper_nibble = (value_to_convert & $f0) >> 4

	if ((upper_nibble >= 0) AND (upper_nibble <= 9)) then
			upper_nibble = upper_nibble + 48
	else
		' The value is between 10 (a) and 15 (f).

		upper_nibble = upper_nibble + 87
	endif 

	lower_nibble = (value_to_convert & $0f)
	
	if ((lower_nibble >= 0) AND (lower_nibble <= 9)) then
			lower_nibble = lower_nibble + 48
	else
		' The value is between 10 (a) and 15 (f).

		lower_nibble = lower_nibble + 87
	endif

return

'************************************************************************
' Routine:        	check_response_delay 

' PreCondition:    	delay_before_responding indicates whether the PIC
'					must delay before responding to a received command.

' Input:           	None

' Output:          	If the delay has timed out, sets network_state = "t"

' Side Effects:    	None

' Overview:        	If a delay before responding is required, check the 
'					timer. If the delay has timed out, enable responding.
                  
' Note:            	
'************************************************************************

check_response_delay:
	if (delay_before_responding = 1) then

		if (INTCON.2 = 1) then
		
			' The delay time has elapsed. Stop the timer and send the response.

			T0CON = 0
			INTCON.2 = 0
			network_state = "t"
		endif
	endif
return

'************************************************************************
' Routine:        	command_read_byte 

' PreCondition:    	A read_byte command has been received.

' Input:           	received_command[] - the command

' Output:          	command_response[] - the response

' Side Effects:    	None

' Overview:       	Processes a read_byte command and prepares to respond.  
                  
' Note:            	
'************************************************************************

command_read_byte:

	' A read_byte command has been received.

	select case received_command[3]

		' Add more cases (locations to read) as needed.
	
		case "b"

			' Read Port B.
	
			value_to_convert = PORTB

			' Convert the value read to ASCII Hex
			
			gosub byte_to_ascii_hex
			
			' Prepare to send the ASCII Hex characters in a response.

			command_response[2] = upper_nibble
			command_response[3] = lower_nibble
			command_response[4] = $0a
		
			gosub prepare_to_respond
				
		case else
	end select
	return

'************************************************************************
' Routine:        	command_write_byte

' PreCondition:    	A write_byte command has been received.

' Input:           	received_command[] - the command

' Output:          	command_response[] - the response

' Side Effects:    	None

' Overview:        	Processes a write_byte command and prepares to respond.  
                  
' Note:            	None
'************************************************************************

command_write_byte:

	select case received_command[3]

		' Add more cases as needed.
	
		case "b"

			' Get the data to write.
			' Convert the received ASCII Hex bytes to a byte value.

			upper_nibble = received_command[4]
			lower_nibble = received_command[5]

			gosub ascii_hex_to_byte

			' Set bit 0 of PortB to match bit 0 in the received byte.

			if ((converted_byte & 1) = 1) then
				high PORTB.0
			else
				low PORTB.0
			endif

			command_response[2] = $0a
			gosub prepare_to_respond

		case else
	end select
return

'************************************************************************
' Routine:        	initialize_serial_buffers 

' PreCondition:    	None

' Input:           	None

' Output:          	None

' Side Effects:    	None

' Overview:        	initalize variables relating to serial data
                  
' Note:            	
'************************************************************************

initialize_serial_buffers:

	command_index = 0
	response_index = 0		
	received_command[0]= 0
return

'************************************************************************
' Routine:        	prepare_to_respond 

' PreCondition:    	The PIC has a response to send.
'					delay_before_responding indicates whether to delay (1)
'					or not (0) before sending the response.

' Input:           	None

' Output:          	network_state is set to "d" or "t"

' Side Effects:    	None

' Overview:        	Prepares to respond to a command.
'					If a delay is needed, sets network_state = "d".
'                 	Otherwise sets network_state = "t".
' Note:            	

'************************************************************************

prepare_to_respond:

	response_index = 0;
	if (delay_before_responding = 1) then 		
		gosub start_response_delay_timer
		network_state = "d"
	else
		network_state = "t"
	endif
return

'************************************************************************
' Routine:        	receive_serial_data 

' PreCondition:    	An open serial port. See DEFINE HSER_* statements

' Input:           	None

' Output:          	received_command[] stores received command bytes
'
' Side Effects:    	None

' Overview:        	Processes received bytes.
                  
' Note:            	None
'************************************************************************

receive_serial_data:

	' Process received bytes.

	if (PIR1.5 = 1) then

		' A byte is available to read.

		if (RCSTA.2 = 1) then
	
			' Framing error. Read RCREG to clear the error
			' but don't use the data.

			hserin [serial_in]
		
		else
	
			' No framing error occurred

			hserin [serial_in]
		
			select case serial_in

				case $0a
	
					' A LF character was received, indicating the end of a command.
		
					gosub respond_to_command

				case $0d
	
					' Ignore a received CR character.

				case COMMAND_START

					' A new command has begun.
					' Initialize the array that holds received bytes.

					received_command[0] = COMMAND_START
					command_index = 1
		
				case else
		
					' A character was received and it's not a LF or CR.
					' The primary node might send extraneous data.
					' If at the end of the array, ignore additional received data.
					
					if (command_index <= MAX_COMMAND_LENGTH) then			

						' Convert characters A-Z to lower case.
			
						if ((serial_in >= "A") and (serial_in <= "Z")) then
							serial_in = serial_in + 32
						endif
			
						' Save the character and increment the position in the array
						' that stores received text.			
		
						received_command[command_index] = serial_in
						command_index = command_index + 1

					endif
			end select		
		endif
	endif
return

'************************************************************************
' Routine:        	respond_to_command 

' PreCondition:    	None

' Input:           	None

' Output:          	None
'
' Side Effects:    	None

' Overview:        	Check for a valid, supported command and
'					if detected, call a routine to handle the command.
                  
' Note:            	None
'************************************************************************

respond_to_command:

	' Every command begins with a COMMAND_START code, the node's address, 
	' and a command code.

	if (received_command[0] = COMMAND_START)  then

		if (received_command[1] = MY_ADDRESS) then

			select case received_command[2]

				case "1"
					gosub command_write_byte

				case "2"
					gosub command_read_byte

				case else
					gosub initialize_serial_buffers

			end select
		else
			gosub initialize_serial_buffers
		endif
	else
		gosub initialize_serial_buffers		
	endif
return

'************************************************************************
' Routine:        	serial_communications 

' PreCondition:    	None

' Input:           	None

' Output:          	None

' Side Effects:    	None

' Overview:        	Implements a state machine for serial communications tasks.
                  
' Note:            	
'************************************************************************

serial_communications:

	select case network_state

		case "r"
			gosub receive_serial_data

		case "d"
			gosub check_response_delay

		case "t"
			gosub transmit_serial_data

		case else

	end select
return

'************************************************************************
' Routine:        	start_response_delay_timer 

' PreCondition:    	None

' Input:           	None

' Output:          	None

' Side Effects:    	Uses TIMER0

' Overview:        	Starts a timer to implement a delay before responding.
                  
' Note:            	
'************************************************************************

start_response_delay_timer:

	' This example sets a delay of around 0.5 sec. assuming FOSC = 4 Mhz.

	' Timer enabled, 16-bit, internal clock, prescaler = 256.

	T0CON = $87

	' Load the timer with F800h. 
	' Both bytes load on a write to TMR0L, so set the value of TMR0H first.

	TMR0H =	$F8
	TMR0L = $00
	
return

'************************************************************************
' Routine:        	transmit_serial_data

' PreCondition:    	An open serial port. See DEFINE HSER_* statements

' Input:           	None

' Output:          	None
'
' Side Effects:    	None

' Overview:        	If a byte is waiting to transmit, send it.
                  
' Note:            	None
'************************************************************************

transmit_serial_data:

		if firmware_driver_enable then
			driver_enable = 1
		endif

		' Wait for the transmit shift register to empty.

		while (TXSTA.1 = 0)
		wend	
		hserout[command_response[response_index]]

		if (command_response[response_index] = $0a) then

			' The entire response has been sent.
			
			if firmware_driver_enable then
				while (TXSTA.1 = 0)
				wend
				driver_enable = 0 
			endif			

			gosub initialize_serial_buffers
			network_state = "r"

		else
			' Prepare to send another byte.

			response_index = response_index + 1
		endif
return







    


















